#include <SoftwareSerial.h>
#include <Rtc_Pcf8563.h>
#include <DHT.h>
#include "settings.h"
#include "button.h"

enum AlarmClock {
  START, CLOCK, MENU, SHOW_TEMP, SHOW_HUM,
  SET_ALARM, SET_TIME, SET_HUM_LIMIT, LUMINOSITY
};


// init uart communication
static const int RX1pin = PD4;
static const int TX1pin = PD3;
static const int RX2pin = PB1;
static const int TX2pin = PB0;

static SoftwareSerial slave1(8, 9);
static SoftwareSerial slave2(3, 4);

// init real-time clock
Rtc_Pcf8563 rtc;
static char currTime[10] = "";
static int currHour = 0;
static int currMin = 0;
static int hour;
static int minute;
static int timeCursor;
static byte luminosity = 100;
static byte gammaLuminosity = 100;
static unsigned long lastTimeRTC = 0;

// init button pins
static const int selRetPin = PE0;
static const int upPin = PD5;
static const int downPin = PD6;
static const Button selRetButton(true, 1 << selRetPin, &PINE);
static const Button upButton(false, 1 << upPin, &PIND);
static const Button downButton(false, 1 << downPin, &PIND);

// init DHT11 sensor
static int dhtPin = 14;
DHT dht(dhtPin, DHT11);
static unsigned long lastTimeReadSensor = 0;
static int humidity;
static int temperature;

// set humidity limit
static int humidityLimit;

// set audio pin
static int audioPin = PC2;

// set mosfet pin
static int mosfetPin = PD7;

static const Settings menu;
static int cursor = menu.getCursor();
static unsigned long lastMillis = 0;
static unsigned long lastSendComm = 0;
static int machineState = CLOCK; 


static char tmpTime[20];
void readRTC() {
  if (millis() - lastTimeRTC > 10000 || lastTimeRTC == 0) {
    hour = rtc.getHour();
    minute = rtc.getMinute();
    strcpy(tmpTime, rtc.formatTime(RTCC_TIME_HM));
    lastTimeRTC = millis();
  }
}

void readHum(unsigned long delayReadSensor) {
  if (millis() - lastTimeReadSensor > delayReadSensor) {
    humidity = dht.readHumidity();
    lastTimeReadSensor = millis();
  }
}

void readTemp(unsigned long delayReadSensor) {
  if (millis() - lastTimeReadSensor > delayReadSensor) {
    temperature = dht.readTemperature();
    lastTimeReadSensor = millis();
  }
}


void sendCommand(char* command) {
  slave1.write(1);
  slave1.write((byte)(gammaLuminosity * 2.5));
  slave1.println(command);
  slave2.write(2);
  slave2.write((byte)(gammaLuminosity * 2.5));
  slave2.println(command);
}


void clockCommand(bool force = false) {
  if (strcmp(currTime, tmpTime) != 0 || force) {
    char command[30];
    memset(command, 0, sizeof(command));
    strcpy(currTime, tmpTime);
    sprintf(command, "C %s:0:0", currTime);
    sendCommand(command);        
  }
}

void menuCommand() {
  char command[30] = "M ";
  char cursorStr[2];
  sprintf(cursorStr, "%d", menu.getCursor());
  strcat(command, cursorStr);
  for (int i = 0; i < 4; i++) {
    char state[4];
    sprintf(state, " %d", menu.getAlarm(i)->getAlarmState());
    strcat(command, state);
  }
  sendCommand(command);
}

void alarmCommand(int tmpCursor) {
  // get current alarm value  
  if (menu.getAlarm(cursor)->getHour() == 24 ||
      menu.getAlarm(cursor)->getMin() == 60) {
    tmpCursor = -1;       
  }

  char currAlarm[5];
  menu.getAlarm(cursor)->getAlarm(currAlarm);
  
  // create and send command
  char command[30] = "A ";
  strcat(command, currAlarm);
  if (tmpCursor == -1) {
    strcat(command, ":2:2");    
  } else if (tmpCursor == 0) {
    strcat(command, ":1:0");
  } else if (tmpCursor == 1) {
    strcat(command, ":0:1");
  }
  sendCommand(command);
}


void setTimeCommand() {
  char command[80] = "I ";
  char time[6];
  sprintf(time, "%d:%d", currHour, currMin);
  strcat(command, time);
  if (timeCursor == 0) {
    strcat(command, ":1:0");
  } else if (timeCursor == 1) {
    strcat(command, ":0:1");
  }
  sendCommand(command);
}

void setHumidityLimitCommand() {
  char command[30];
  sprintf(command,  "L %d", humidityLimit);
  sendCommand(command);        
}

void calculateGamma() {
  gammaLuminosity = pow(luminosity  * 0.01, 2.8) * 100;
}

void setLuminosityLimit() {
  char command[30];
  calculateGamma();
  sprintf(command,  "L %d", (int)luminosity);
  sendCommand(command);        
}

bool tmpBit = false;
void checkAlarm() {
  for (int i = 0; i < 4; i++) {   
    if (menu.getAlarm(i)->getHour() == hour &&
        menu.getAlarm(i)->getMin() == minute &&
        menu.getAlarm(i)->getAlarmState() == 1) {
      if (lastMillis != millis()) {        
        if (tmpBit = !tmpBit) {
          PORTC |= (1 << audioPin);
        } else {
          PORTC &= ~(1 << audioPin);        
        }
        lastMillis = millis();     
      }  
    } 
  }
}

void checkHumidity() {
  if (humidity > humidityLimit) {
    PORTD |= (1 << mosfetPin);
  } else {
    PORTD &= ~(1 << mosfetPin);
  }
}


void setup() {
  // activate the pull-up resistance for buttons
  DDRE &= ~(1 << selRetPin);
  PORTE |= (1 << selRetPin);
  DDRD &= ~(1 << upPin);
  PORTD |= (1 << upPin);
  DDRD &= ~(1 << downPin);
  PORTD |= (1 << downPin);
    
  // set the baud rate for uart communication
  slave1.begin(9600);
  slave2.begin(9600);

  // activate dht11 sensor  
  dht.begin();

  // read luminosity
  luminosity = EEPROM.read(201);
  if (luminosity < 25 || luminosity > 99) {
    luminosity = 70; // default value
  }
  calculateGamma();

  // set audioPin as output
  DDRC |= (1 << audioPin);

  // set mosfetPin as output
  DDRD |= (1 << mosfetPin);

  // get humidity limit from EEPROM
  byte valHumidityLimit = EEPROM.read(200);    
  humidityLimit = (valHumidityLimit == 0xFF) ? 99 : (int)valHumidityLimit;

  machineState = START;
}

void loop() {
  readRTC();
  checkAlarm();
  checkHumidity();

  selRetButton.update();
  upButton.update();
  downButton.update();
  
  switch(machineState) {
    case START:
      readRTC();
      clockCommand(true);
      machineState = CLOCK;
      break;      
    
    case CLOCK:
      if (selRetButton.isShortPress()) {
        menuCommand();
        machineState = MENU;
      } else {        
        clockCommand();
        readHum(20000);
        readTemp(20000);   
      }
      if (selRetButton.isLongPress()) {
        setLuminosityLimit();        
        machineState = LUMINOSITY;        
      }      
      break;
      
    case MENU:
      if (downButton.isShortPress()) {
        menu.incCursor();
        menuCommand();
      } else if (upButton.isShortPress()) {
        menu.decCursor();
        menuCommand();
      } else if (selRetButton.isShortPress()) {
        cursor = menu.getCursor();
        if (cursor >= 0 && cursor < 4) {
          // init cursor in alarm setup
          menu.getAlarm(cursor)->initAlarm();
          alarmCommand(0);
          machineState = SET_ALARM;
        } else if (cursor == 4) {
          timeCursor = 0;
          currHour = rtc.getHour();
          currMin = rtc.getMinute();
          setTimeCommand();
          machineState = SET_TIME;
        } else if (cursor == 5) {
          setHumidityLimitCommand();
          machineState = SET_HUM_LIMIT;
        } else if (cursor == 6) {
          char command[30];
          sprintf(command,  "H %d", humidity);
          sendCommand(command);
          lastSendComm = millis();
          machineState = SHOW_HUM;
        } else if (cursor == 7) {
          char command[30];
          sprintf(command,  "T %d", temperature);
          sendCommand(command);
          lastSendComm = millis();
          machineState = SHOW_TEMP;
        }
      } else if (selRetButton.isLongPress()) {
        menu.resetCursor();
        clockCommand(true);
        machineState = CLOCK;
      }
      break;

    case SET_ALARM: {
      int tmpCursor = menu.getAlarm(cursor)->getAlCursor();
      if (downButton.isShortPress()) {
        if (tmpCursor == 0) {
          menu.getAlarm(cursor)->incHour();
          alarmCommand(tmpCursor);
        } else if (tmpCursor == 1) {
          menu.getAlarm(cursor)->incMin();
          alarmCommand(tmpCursor);
        }
      } else if (upButton.isShortPress()) {
        if (tmpCursor == 0) {
          menu.getAlarm(cursor)->decHour();
          alarmCommand(tmpCursor);
        } else if (tmpCursor == 1) {
          menu.getAlarm(cursor)->decMin();
          alarmCommand(tmpCursor);
        }
      } else if (selRetButton.isShortPress()) {
        if (menu.getAlarm(cursor)->getHour() == 24 ||
            menu.getAlarm(cursor)->getMin() == 60) {
          menu.getAlarm(cursor)->deleteAlarm();
          menuCommand();
          machineState = MENU;
        } else if (tmpCursor == 1) {
          menu.getAlarm(cursor)->setAlarm();
          menuCommand();
          machineState = MENU;
        } else {                              
          menu.getAlarm(cursor)->incAlCursor();
          alarmCommand(1);
        }
      } else if (selRetButton.isLongPress()) {
        menuCommand();
        machineState = MENU;
      }
      break;
    }      

    case SET_TIME:
      if (downButton.isShortPress()) {
        if (timeCursor == 0) {
          if (currHour == 23) {
            currHour = 0;
          } else {
            currHour++;            
          }          
        } else if (timeCursor == 1) {
          if (currMin == 59) {
            currMin = 0;
          } else {
            currMin++;            
          }          
        }
        setTimeCommand();
      } else if (upButton.isShortPress()) {
        if (timeCursor == 0) {
          if (currHour == 0) {
            currHour = 23;
          } else {
            currHour--;            
          }          
        } else if (timeCursor == 1) {
          if (currMin == 0) {
            currMin = 59;
          } else {
            currMin--;            
          }          
        }
        setTimeCommand();
      } else if (selRetButton.isShortPress()) {
        if (timeCursor == 1) {
          rtc.initClock();
          rtc.setTime(currHour, currMin, 0);
          menuCommand();
          machineState = MENU;
        } else {
          timeCursor++;
          setTimeCommand();                                                 
        }
      } else if (selRetButton.isLongPress()) {
        menuCommand();
        machineState = MENU;
      }
      break;

    case SET_HUM_LIMIT:
      if (downButton.isShortPress()) {
        if (humidityLimit < 99) {
          humidityLimit++;  
          setHumidityLimitCommand();        
        }  
      } else if (upButton.isShortPress()) {
        if (humidityLimit > 0) {
          humidityLimit--;
          setHumidityLimitCommand();
        }
      } else if (selRetButton.isShortPress()) {
        EEPROM.write(200, humidityLimit & 0xFF);
        menuCommand();
        machineState = MENU;
      } else if (selRetButton.isLongPress()) {
        menuCommand();
        machineState = MENU;
      }
      break;

    case SHOW_HUM:
      if (selRetButton.isLongPress()) {
        menuCommand();
        machineState = MENU;
      } else {
        readHum(5000);
        if (millis() - lastSendComm > 5000) {
          char command[30];
          sprintf(command,  "H %d", humidity);
          sendCommand(command);
          lastSendComm = millis();          
        }
      }
      break;

    case SHOW_TEMP:
      if (selRetButton.isLongPress()) {
        menuCommand();
        machineState = MENU;
      } else {
        readTemp(5000);
        if (millis() - lastSendComm > 5000) {
          char command[30];
          sprintf(command,  "T %d", temperature);
          sendCommand(command);
          lastSendComm = millis();
        }
      }
      break;
    case LUMINOSITY:
      if (downButton.isShortPress()) {
        if (luminosity < 99) {
          luminosity++;  
          setLuminosityLimit();        
        }  
      }
      if (upButton.isShortPress()) {
        if (luminosity > 25) {
          luminosity--;
          setLuminosityLimit();
        }
      }
      if (selRetButton.isLongPress() || selRetButton.isShortPress()) {
        EEPROM.write(201, luminosity);
        clockCommand(true);
        machineState = CLOCK;
      }
      break;
  }
}
