#include <SoftwareSerial.h>
#include <avr/pgmspace.h>
#include "standard_matrices_draw.h"

#define DIGIT_0 0
#define DIGIT_1 1
#define DIGIT_2 2
#define DIGIT_3 3
#define DIGIT_4 4
#define DIGIT_5 5
#define DIGIT_6 6
#define DIGIT_7 7lu
#define DIGIT_8 8
#define DIGIT_9 9
#define SETTINGS_1 10
#define SETTINGS_2 11
#define CELSIUS_1 12
#define CELSIUS_2 13
#define PERCENT_1 14
#define PERCENT_2 15
#define CURSOR 16
#define PLUS 17
#define MINUS 18
#define FLAGS 19
#define LINE 20

#define LEFT_SLAVE 1
#define RIGHT_SLAVE 2

// slave 1 - left or 2 - right
static int noSlave;

static const int dataPin = PC0;
static const int clockPin = PC1;
static const int latchPin = PC2;
static const int outenPin = 5;

static uint8_t frame_buffer[16][16];

static const int RXpin = PD3;
static const int TXpin = PD4;
static SoftwareSerial serial(3, 4);

static char command[30] = "";
static int count = 0;
bool sendingLuminosity = false;


void write_in_frame_buffer(uint8_t idx, uint8_t pos_x, uint8_t pos_y) {
  // get the dimensions of matrix to draw from flash memory
  uint8_t dim_x = pgm_read_byte(matrix_table[idx]);
  uint8_t dim_y = pgm_read_byte(matrix_table[idx] + 1);
  // write in frame buffer from (pos_x, pos_y)
  for (uint8_t i = pos_x; i < pos_x + dim_x; i++) {
    uint8_t mask = 0b10000000;
    uint8_t offset = i - pos_x + 2;
    uint8_t line_to_draw = pgm_read_byte(matrix_table[idx] + offset);
    
    for (uint8_t j = pos_y; j < pos_y + dim_y; j++) {
      if ((line_to_draw & mask)) {
        frame_buffer[i][j] = 1;
      } else {
        frame_buffer[i][j] = 0;
      }
      mask = mask >> 1;
    }
  }
}


// Clear the frame buffer
void clear_frame_buffer() {
  for (uint8_t i = 0; i < 16; i++) {
    for (uint8_t j = 0; j < 16; j++) {
      frame_buffer[i][j] = 0;
    }
  }
}


// ShiftOut 256-bit
void shiftOut_256(uint8_t dataPin, uint8_t clockPin) {
  PORTC &= ~(1 << latchPin);
  for (uint8_t k = 0; k < 2; k++) {
    for (uint8_t i = 0; i < 16; i++) {
      for (uint8_t j = k * 8; j < (k + 1) * 8; j++) {
        if (frame_buffer[i][j] == 0) {
          // digitalWrite(dataPin, LOW)
          PORTC &= ~(1 << dataPin);
        } else {
          // digitalWrite(dataPin, HIGH)
          PORTC |= (1 << dataPin);
        }

        // digitalWrite(clockPin, HIGH)
        // digitalWrite(clockPin, LOW)
        PORTC |= (1 << clockPin);
        PORTC &= ~(1 << clockPin);
      }
    }
  }
  PORTC |= (1 << latchPin);
  PORTC &= ~(1 << latchPin);
}

// Draw number
void draw_number(int value, uint8_t pos_x, uint8_t pos_y) {
  write_in_frame_buffer(value / 10, pos_x, pos_y);
  write_in_frame_buffer(value % 10, pos_x, pos_y + 8);
}

// Draw -- --
void draw_no_numbers() {
  write_in_frame_buffer(LINE, 7, 0);
  write_in_frame_buffer(LINE, 7, 8);
}

// Draw flags
void draw_flags() {
  write_in_frame_buffer(FLAGS, 0, 0);
  write_in_frame_buffer(FLAGS, 0, 8);
  write_in_frame_buffer(FLAGS, 15, 0);
  write_in_frame_buffer(FLAGS, 15, 8);
}

// Draw clock
void draw_clock(char* clock_value) {
  char delim[2] = ":";
  int hour, min;
  
  char* ptr = strtok(clock_value, delim);
  hour = atoi(ptr);
  ptr = strtok(NULL, delim);
  min = atoi(ptr);
  int flagL, flagR;
  if (ptr = strtok(NULL, delim)) {
    flagL = atoi(ptr);
  }
  if (ptr = strtok(NULL, delim)) {
    flagR = atoi(ptr);
  }

  if (noSlave == LEFT_SLAVE) {
    if (flagL == 2) {
      draw_no_numbers();
    } else {
      draw_number(hour, 2, 0);
      if (flagL == 1) {
        draw_flags();
      }   
    }
  } else {
    if (flagR == 2) {
      draw_no_numbers();
    } else {
      draw_number(min, 2, 0);
      if (flagR == 1) {
        draw_flags();
      }   
    }    
  }
}


// Draw settings
void draw_settings(char* val) {
  char delim[2] = " ";
  int cursor, alarm[4], count = 0;
  
  char* ptr = strtok(val, delim);
  cursor = atoi(ptr);

  ptr = strtok(NULL, delim);
  while (ptr != NULL) {
    // 0 - alarm to be set
    // 1 - alarm set
    alarm[count] = atoi(ptr);
    count++;        
    ptr = strtok(NULL, delim);  
  }

  if (noSlave == LEFT_SLAVE) {
    write_in_frame_buffer(SETTINGS_1, 0, 0);
    if (cursor >= 0 && cursor < 4) {
      uint8_t pos_x = 4 * cursor + 1;
      write_in_frame_buffer(CURSOR, pos_x, 8);      
    }
    for (int i = 0; i < 4; i++) {
      uint8_t pos_x = 4 * i + 1;
      uint8_t pos_y = 3;      
      if (alarm[i] == 0) {                
        write_in_frame_buffer(PLUS, pos_x, pos_y);
      } else {
        write_in_frame_buffer(MINUS, pos_x, pos_y);        
      }
    }
  } else {
    write_in_frame_buffer(SETTINGS_2, 0, 0);
    if (cursor >= 4 && cursor < 8) {
      uint8_t pos_x = 4 * (cursor - 4) + 1;
      write_in_frame_buffer(CURSOR, pos_x, 8);      
    }    
  }
}


// Draw temperature
void draw_temperature(char* temp_value) {
  int temperature = atoi(temp_value);
  
  if (noSlave == LEFT_SLAVE) {
    if (temperature > 9) {
      draw_number(temperature, 2, 0);
    } else {
      write_in_frame_buffer(temperature, 2, 4);
    }
  } else {
    write_in_frame_buffer(CELSIUS_1, 2, 0);
    write_in_frame_buffer(CELSIUS_2, 2, 8);
  }
}


// Draw humidity / luminosity
void draw_humidity(char* hum_value) {
  int humidity = atoi(hum_value);
  
  if (noSlave == LEFT_SLAVE) {
    if (humidity > 9) {
      draw_number(humidity, 2, 0);
    } else {
      write_in_frame_buffer(humidity, 2, 4);
    }
  } else {
    write_in_frame_buffer(PERCENT_1, 2, 0);
    write_in_frame_buffer(PERCENT_2, 2, 8);
  }
}



void setup() {
  // set the shift register's pins to output
  DDRC |= (1 << dataPin);
  DDRC |= (1 << clockPin);
  DDRC |= (1 << latchPin);
  pinMode(outenPin, OUTPUT);
  analogWrite(outenPin, 70);
  
  // set baud rate for uart transmission
  Serial.begin(115200);
  serial.begin(9600);  
}


void loop() {
  if (serial.available()) {
    char tmp;
    tmp = serial.read();
    if (sendingLuminosity) {
      analogWrite(outenPin, 255 - tmp);
      sendingLuminosity = false;
      return;
    }
    if (tmp == 1 || tmp == 2) {
      noSlave = tmp;
      sendingLuminosity = true;
      return;
    }
    if (tmp != '\n') {
      command[count] = tmp;
      count++;
    } else {
      command[count] = '\0';
      Serial.println(command);
      
      clear_frame_buffer();
      switch (command[0]) {
        case 'C':
        case 'A':
        case 'I':
          draw_clock(command + 2);
          break;
        case 'M':
          draw_settings(command + 2);
          break;
        case 'L':
        case 'H':
          draw_humidity(command + 2);
          break;
        case 'T':
          draw_temperature(command + 2);
          break;  
      }
      shiftOut_256(dataPin, clockPin);   
      
      count = 0; 
    }
  }
}